import logging
import traceback
import uuid
from typing import Any
from typing import List
from typing import Optional

from openlineage.airflow.extractors.base import TaskMetadata

# TODO submit bug fix to OL for:
from openlineage.airflow.extractors.bigquery_extractor import BigQueryExtractor
from openlineage.airflow.utils import get_job_name
from openlineage.client.facet import SqlJobFacet

# from . import bigquery_extractor
from openlineage.common.provider.bigquery import BigQueryDatasetsProvider
from openlineage.common.provider.bigquery import BigQueryErrorRunFacet

from . import alvin_vars
from . import util

log = logging.getLogger(__name__)


def _extract_bq_eq(operator):
    data_dict = {}
    # connection/auth context
    data_dict["project_id"] = operator.hook.project_id
    # # TODO figure out how to parse to string XComArg in "operator.output"
    print(f"{type(operator).__name__} parsed context: {data_dict}")
    return data_dict


class AlvinBQExtractor(BigQueryExtractor):
    """
    https://github.com/OpenLineage/OpenLineage/tree/main/integration/airflow#custom-extractors
    OPENLINEAGE_EXTRACTOR_BigQueryOperator=alvin_lineage.alvin_bq_extractor.AlvinBQExtractor
    """

    def __init__(self, operator):
        super().__init__(operator)
        util.print_alvin_environment(AlvinBQExtractor.__name__)

    @classmethod
    def get_operator_classnames(cls) -> List[str]:
        return alvin_vars.BQ_OPERATORS

    def extract_on_complete(self, task_instance) -> Optional[TaskMetadata]:
        log.debug(f"extract_on_complete({task_instance})")
        context = self.parse_sql_context()

        try:
            bigquery_job_id = self._get_xcom_bigquery_job_id(task_instance)
            if bigquery_job_id is None:
                raise Exception(
                    "Xcom could not resolve BigQuery job id." + "Job may have failed."
                )
        except Exception as e:
            log.error(
                f"Cannot retrieve job details from BigQuery.Client. {e}", exc_info=True
            )
            return TaskMetadata(
                name=get_job_name(task=self.operator),
                run_facets={
                    "bigQuery_error": BigQueryErrorRunFacet(
                        clientError=f"{e}: {traceback.format_exc()}",
                        parserError=context.parser_error,
                    )
                },
            )

        # BEGIN bugfix to OL
        bq_client = self.operator.hook.get_client()
        log.info(f"Using this bq_client: {bq_client}")
        stats = BigQueryDatasetsProvider(bq_client).get_facets(bigquery_job_id)
        # END bugfix to OL
        inputs = stats.inputs
        output = stats.output
        run_facets = stats.run_facets

        return TaskMetadata(
            name=get_job_name(task=self.operator),
            inputs=[ds.to_openlineage_dataset() for ds in inputs],
            outputs=[output.to_openlineage_dataset()] if output else [],
            run_facets=run_facets,
            job_facets={
                # BEGIN custom Alvin stuff
                "alvin": {
                    "generated_job_id": f"{self.operator.dag_id}.{self.operator.task_id}.{uuid.uuid4()}",
                    "task_ctx": self._extract_stuff(),
                },
                # END custom Alvin stuff
                "sql": SqlJobFacet(context.sql),
            },
        )

    def _extract_stuff(self) -> Any:
        if self.operator.task_type == "BigQueryExecuteQueryOperator":
            return _extract_bq_eq(self.operator)
        log.warning(
            f"Cannot find a project ID for this Task Type: {self.operator.task_type}"
        )
        return None

    def extract(self, task_instance) -> Optional[TaskMetadata]:
        return self.extract_on_complete(task_instance)
