import logging
import sys
from contextlib import closing

import gorilla
from airflow.contrib.hooks import snowflake_hook
from airflow.contrib.operators import snowflake_operator
from past.builtins import basestring

log = logging.getLogger(__name__)

"""
https://gorilla.readthedocs.io/en/latest/tutorial.html#tutorial
"""

GORILLA_SETTINGS = gorilla.Settings(
    allow_hit=True,
    # store_hit=True,
)


@gorilla.patches(snowflake_hook.SnowflakeHook, settings=GORILLA_SETTINGS)
class AlvinSnowflakeHook(object):
    @gorilla.name("better_run")
    def run(self, sql, autocommit=False, parameters=None):
        """
        Runs a command or a list of commands. Pass a list of sql
        statements to the sql parameter to get them to execute
        sequentially

        :param sql: the sql statement to be executed (str) or a list of
            sql statements to execute
        :type sql: str or list
        :param autocommit: What to set the connection's autocommit setting to
            before executing the query.
        :type autocommit: bool
        :param parameters: The parameters to render the SQL query with.
        :type parameters: mapping or iterable
        """
        # Alvin: make sure we reset the SF Query IDs list
        self.sfqids = []
        if isinstance(sql, basestring):
            sql = [sql]

        with closing(self.get_conn()) as conn:
            if self.supports_autocommit:
                self.set_autocommit(conn, autocommit)

            with closing(conn.cursor()) as cur:
                for s in sql:
                    if sys.version_info[0] < 3:
                        s = s.encode("utf-8")
                    if parameters is not None:
                        self.log.info("{} with parameters {}".format(s, parameters))
                        cur.execute(s, parameters)
                    else:
                        self.log.info(s)
                        cur.execute(s)
                    # Alvin: overriding this just to collect the sfqids!
                    curr_sfqid = cur.sfqid
                    self.sfqids.append(curr_sfqid)
                    log.warning(
                        f"(AlvinSnowflakeHook): collected SF Query ID: '{curr_sfqid}' for Snowflake SQL:\n{s}\n"
                    )
                log.warning(f"(AlvinSnowflakeHook): all the sfqids are: {self.sfqids}")

            # If autocommit was set to False for db that supports autocommit,
            # or if db does not supports autocommit, we do a manual commit.
            if not self.get_autocommit(conn):
                conn.commit()


@gorilla.patches(snowflake_operator.SnowflakeOperator, settings=GORILLA_SETTINGS)
class AlvinSnowflakeOperator:
    @gorilla.name("better_execute")
    def execute(self, context):
        # Alvin: make sure we reset the SF Query IDs list
        self.sfqids = []
        self.log.info("Executing: %s", self.sql)
        hook = self.get_hook()
        hook.run(self.sql, autocommit=self.autocommit, parameters=self.parameters)
        self.sfqids = hook.sfqids
        log.warning(
            f"(AlvinSnowflakeOperator): these are the collected Snowflake query IDs from the hook: {self.sfqids}"
        )
