from typing import Optional
from typing import Union

from airflow.lineage.backend import LineageBackend
from airflow.version import version as AIRFLOW_VERSION
from openlineage.lineage_backend import Backend
from pkg_resources import parse_version

from . import util

"""
from this doc we see: GCP
https://cloud.google.com/composer/docs/concepts/versioning/composer-versioning-overview?hl=en_US&_ga=2.61917484.-1649017489.1639472707

Cloud Composer 1:
Airflow 1.10.* and Airflow 2

Cloud Composer 2
Airflow 2

AWS: 
https://docs.aws.amazon.com/mwaa/latest/userguide/airflow-versions.html
v1.10.12, v2.0.2 and v2.2.2

Astronomer?
https://docs.astronomer.io/enterprise/manage-airflow-versions/
"""
# https://docs.aws.amazon.com/mwaa/latest/userguide/airflow-versions.html

"""
ahhh, we can simplify this a bit:
https://github.com/apache/airflow/blob/a2abf66315/airflow/models/__init__.py <-- they are all here 
"""

v = parse_version(AIRFLOW_VERSION)
if v <= parse_version("2.2.3") and v >= parse_version("2.0.1"):
    """
    2.2.3 - 2.0.1
    https://github.com/apache/airflow/blob/2.2.3/airflow/models/taskinstance.py
    https://github.com/apache/airflow/blob/2.0.2/airflow/models/taskinstance.py

    from airflow.utils.session import provide_session
    from airflow.models.taskinstance import TaskInstance
    @provide_session
    def handle_failure(
        self,
        error: Union[str, Exception],
        test_mode: Optional[bool] = None,
        force_fail: bool = False,
        error_file: Optional[str] = None,
        session=None,
    ) -> None:
    """
    from airflow.models.taskinstance import TaskInstance

    TaskInstance.handle_failure_wrapped = TaskInstance.handle_failure

    def new_handle_failure(
        self,
        error: Union[str, Exception],
        test_mode: Optional[bool] = None,
        force_fail: bool = False,
        error_file: Optional[str] = None,
        session=None,
    ):
        print("running wrapped failure handler")
        self.handle_failure_wrapped(error, test_mode, force_fail)
        print("done with original failure handler, phoning home")
        print("now, I will call home")
        print("done")

    TaskInstance.handle_failure = new_handle_failure

    """
    update_state: https://github.com/apache/airflow/blob/a2abf66315/airflow/models/dagrun.py
    https://github.com/apache/airflow/blob/1.8.2/airflow/models.py
    """
    from airflow.models.dagrun import DagRun

    DagRun.update_state_wrapped = DagRun.update_state

    def new_update_state(self, session=None, execute_callbacks=None):
        print("running wrapped dagrun update state handler")
        res = self.update_state_wrapped(session, execute_callbacks)
        print("done wrapped dagrun update state handler")
        print(f"State is {self._state}")
        print(self)
        return res

    DagRun.update_state = new_update_state

if v <= parse_version("2.0.0") and v >= parse_version("1.10.11"):
    """
    2.0.0 - 1.10.11
    https://github.com/apache/airflow/blob/2.0.0/airflow/models/taskinstance.py
    https://github.com/apache/airflow/blob/1.10.11/airflow/models/taskinstance.py

    from airflow.models.taskinstance import TaskInstance
    2.0.0: from airflow.utils.session import provide_session
    1.10.11: from airflow.utils.db import provide_session

    @provide_session
    def handle_failure(self, error, test_mode=None, context=None, force_fail=False, session=None):
    """
    pass
if v <= parse_version("1.10.10") and v >= parse_version("1.10.4"):
    """
    1.10.10 - 1.10.4
    https://github.com/apache/airflow/blob/1.10.10/airflow/models/taskinstance.py
    https://github.com/apache/airflow/blob/1.10.4/airflow/models/taskinstance.py


    from airflow.models.taskinstance import TaskInstance
    from airflow.utils.db import provide_session

    @provide_session
    def handle_failure(self, error, test_mode=None, context=None, session=None):
    """
    pass
if v <= parse_version("1.10.3") and v >= parse_version("1.10.0"):
    """
    1.10.2 - 1.10.0
    https://github.com/apache/airflow/blob/1.10.2/airflow/models.py
    https://github.com/apache/airflow/blob/1.10.0/airflow/models.py
    https://github.com/apache/airflow/blob/1.10.3/airflow/models/__init__.py


    from airflow.models import TaskInstance
    from airflow.utils.db import provide_session

    @provide_session
    def handle_failure(self, error, test_mode=None, context=None, session=None):
    -------------------------------------------------------------------------------
    """
    pass
if v <= parse_version("1.9.0") and v >= parse_version("1.6.0"):
    """
    1.9.0 - 1.6.0
    https://github.com/apache/airflow/blob/1.9.0/airflow/models.py
    https://github.com/apache/airflow/blob/1.6.0/airflow/models.py

    from airflow.models import TaskInstance
    def handle_failure(self, error, test_mode=None, context=None):
    """
    pass
if v <= parse_version("1.5.0") and v >= parse_version("1.3.0"):
    """
    1.5.0 - 1.3.0
    https://github.com/apache/airflow/blob/1.5.0/airflow/models.py
    https://github.com/apache/airflow/blob/1.3.0/airflow/models.py

    from airflow.models import TaskInstance
    def handle_failure(self, error, test_mode, context):
    """
    pass
if v <= parse_version("1.2.0"):
    """
    1.2.0 -
    https://github.com/apache/airflow/blob/1.2.0/airflow/models.py

    1.2.0 and below there is no nothing, but seeing latest commit is at (Latest commit db23691 on 1 Jul 2015)
    we might get away with this
    """
    pass
# Code below is for monkey patching airflow task (and potentially DAG) methods so we can inject callbacks for failure or success.
# we may need to look through different versions of airflow to support different versions, but I guess it's easier to just start with
# airflow 2 support and then work our way backwards.

# NB it does seem that handle_failure is the most consistent thing avaiable. lets see what params it has and how we can easily override it.

# it seems that all tasks have the run method and provide session - if we override the run method, maybe we can hook into the lifecycle that way
# the latest version of airflow is 2.2.3, we are hooking into private methods no part of public API so probably make sense to pin it
# if parse_version(AIRFLOW_VERSION) >= parse_version("2.0.2") and parse_version(
#     AIRFLOW_VERSION
# ) <= parse_version("2.2.3"):
#     # NICE! here it is working sample code of how we can listen in to events in the airflow lifecycle :D
#     # for the finished https://github.com/apache/airflow/blob/1a8a897120762692ca98ac5ce4da881678c073aa/airflow/models/taskinstance.py#L1571
#     # https://github.com/apache/airflow/blob/1a8a897120762692ca98ac5ce4da881678c073aa/airflow/models/taskinstance.py#L1563 before task is starting -
#     # pretty cool stuff
#     try:
#         from airflow.models.taskinstance import TaskInstance

#         TaskInstance._run_finished_callback_wrapped = (
#             TaskInstance._run_finished_callback
#         )
#         # run finished callback is available from 2.0.2 onwards
#         def new_finished_callback(self, error=None):
#             print("running wrapped callback")
#             self._run_finished_callback_wrapped(error)
#             print("done with original callback, phoning home")
#             print("now, I will call home")
#             print("done")

#         TaskInstance._run_finished_callback = new_finished_callback
#     except ImportError as e:
#         print(
#             "Could not do import 'from airflow.models.taskinstance import TaskInstance'"
#         )
#     except Exception as e:
#         print(f"Could monkey patch TaskInstance: {e}")
# elif parse_version(AIRFLOW_VERSION) >= parse_version("1.10.4"):
#     # https://github.com/apache/airflow/tree/1.10.11
#     # this is the first one where models are separated from the models file into a module
#     # https://github.com/apache/airflow/blob/1.10.11/airflow/models/taskinstance.py

#     # https://github.com/apache/airflow/blob/1.10.3/airflow/models/taskinstance.py this doesn not exist, so it came in 1.10.4
#     pass
# elif parse_version(AIRFLOW_VERSION) >= parse_version("1.3.0"):
#     # in these verrsions there is just the airflow.models.TaskInstance vs airflow.models.taskinstance.TaskInstance
#     # between 1.3.0 and 2.0.0
#     # handle failure https://github.com/apache/airflow/blob/5786dcdc392f7a2649f398353a0beebef01c428e/airflow/models/taskinstance.py#L1144
#     # run raw task: https://github.com/apache/airflow/blob/5786dcdc392f7a2649f398353a0beebef01c428e/airflow/models/taskinstance.py#L906
#     # Decorators: https://github.com/apache/airflow/blob/5786dcdc392f7a2649f398353a0beebef01c428e/airflow/utils/db.py#L53

#     # in then handle_failure override - we need to check the state of the task if it is fully failed OR if it is being retried
#     # by running the replaced method before.

#     pass
# else:
#     # below 1.3.0 - we nee to have a look at the run method possibly
#     # https://github.com/apache/airflow/blob/46e56db2f6e572278069cc8862cd0bca81526521/airflow/models.py#L740
#     pass

# original_finished_callback = TaskInstance._run_finished_callback


API_KEY_HEADER = "X-API-KEY"  # from `alvin/backend/app/auth/api_key.py` TODO make this coming from an Alvin core lib to be imported


def _add_auth_aws_flavor(self, api_key: str):
    self.session.headers.update({API_KEY_HEADER: api_key})
    print(
        f"Alvin patched the OL API KEY header as {API_KEY_HEADER}={self.session.headers.get(API_KEY_HEADER)}"
    )


def patch_open_lineage_api_key_header():
    from openlineage.client import OpenLineageClient

    OpenLineageClient._add_auth_old = OpenLineageClient._add_auth

    OpenLineageClient._add_auth = _add_auth_aws_flavor


patch_open_lineage_api_key_header()


class AlvinBackend(Backend):
    def __init__(self):
        super().__init__()
        util.setup_environment_config()


class AlvinLineageBackend(LineageBackend):
    # Airflow 1.10 uses send_lineage as staticmethod, so just construct class
    # instance on first use and delegate calls to it
    backend: AlvinBackend = None

    @classmethod
    def send_lineage(cls, *args, **kwargs):

        print("calling send lineage with args and kwargs")
        if not cls.backend:
            print("Lineage backend is none, creating it")
            cls.backend = AlvinBackend()
        return cls.backend.send_lineage(*args, **kwargs)
